from sklearn.base import TransformerMixin

class FeatureLagger(TransformerMixin):
    '''
        Adds a lagged version of "feature_name", using "keys" to join to the original data
    '''
    def __init__(self, feature_name, date_id="date_block_num", keys=["shop_id", "item_id"],
                 period_lookback = 1,
                 use_last:bool = False,
                 min_items:int = 0,
                 use_test_data:bool=False,
                 fill_test_data_value:float=None,
                 verbose:bool=False # refactor: use a logger
                ):
        self.feature_name = feature_name
        self.date_id = date_id
        self.keys = keys
        self.period_lookback = period_lookback # aka lag
        self.use_last = use_last
        self.use_test_data = use_test_data
        self.fill_test_data_value = fill_test_data_value
        self.train_data = None
        self.lagged_name = f"{self.feature_name}_lag_{period_lookback}"
        self.verbose = verbose
    def fit(self, X, y=None):
        self.train_data = X[[self.date_id, self.feature_name] + self.keys]
        id_max_train = int(self.train_data[self.date_id].max())
        
        self.train_data.loc[:, self.date_id] = self.train_data.loc[:, self.date_id] + self.period_lookback
        self.train_data.rename(columns={self.feature_name: self.lagged_name}, inplace=True)

        return self
    
    def transform(self, X):
        x_len = X.shape[0]
        merge_keys = [self.date_id] + self.keys
        
        if self.verbose:
            print(f"Merge keys: {merge_keys}")
        X = X.merge(self.train_data, on=merge_keys, how="left")
        

        if self.fill_test_data_value is not None:
            X[self.lagged_name] = X[self.lagged_name].fillna(self.fill_test_data_value)
        
        assert x_len == X.shape[0], print(f"initial len: {x_len}, final len: {X.shape[0]}")
        
        return X


import numpy as np
from sklearn.utils import indexable
from sklearn.utils.validation import _num_samples
from sklearn.model_selection._split import _BaseKFold


class GroupTimeSeriesSplit(_BaseKFold):
    """
    TODO: refactor.
    * This is widely used on kaggle, also in its Purged version.
    * See also this stackoverflow question: https://stackoverflow.com/questions/51963713/cross-validation-for-grouped-time-series-panel-data

    Time Series cross-validator for a variable number of observations within the time 
    unit. In the kth split, it returns first k folds as train set and the (k+1)th fold 
    as test set. Indices can be grouped so that they enter the CV fold together.

    Parameters
    ----------
    n_splits : int, default=5
        Number of splits. Must be at least 2.
    max_train_size : int, default=None
        Maximum size for a single training set.
    """
    def __init__(self, n_splits=5, *, max_train_size=None):
        super().__init__(n_splits, shuffle=False, random_state=None)
        self.max_train_size = max_train_size

    def split(self, X, y=None, groups=None):
        """
        Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and n_features is 
            the number of features.
        y : array-like of shape (n_samples,)
            Always ignored, exists for compatibility.
        groups : array-like of shape (n_samples,)
            Group labels for the samples used while splitting the dataset into 
            train/test set.
            Most often just a time feature.

        Yields
        -------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """
        n_splits = self.n_splits
        X, y, groups = indexable(X, y, groups)
        n_samples = _num_samples(X)
        n_folds = n_splits + 1
        indices = np.arange(n_samples)
        group_counts = np.unique(groups, return_counts=True)[1]
        groups = np.split(indices, np.cumsum(group_counts)[:-1])
        n_groups = _num_samples(groups)
        if n_folds > n_groups:
            raise ValueError(
                ("Cannot have number of folds ={0} greater"
                 " than the number of groups: {1}.").format(n_folds, n_groups))
        test_size = (n_groups // n_folds)
        test_starts = range(test_size + n_groups % n_folds,
                            n_groups, test_size)
        for test_start in test_starts:
            if self.max_train_size:
                train_start = np.searchsorted(
                    np.cumsum(
                        group_counts[:test_start][::-1])[::-1] < self.max_train_size + 1, 
                        True)
                yield (np.concatenate(groups[train_start:test_start]),
                       np.concatenate(groups[test_start:test_start + test_size]))
            else:
                yield (np.concatenate(groups[:test_start]),
                       np.concatenate(groups[test_start:test_start + test_size]))